#pragma once
// vim: set ts=2 sw=2 et:

#ifdef _PROFILE

#include <stack>
#include <string>
#include <chrono>
#include <numeric>
#include <sstream>

struct Timestamp;
typedef std::stack<Timestamp *> TimestampStack;
typedef std::chrono::nanoseconds TimestampUnit;
typedef std::chrono::high_resolution_clock TimestampClock;
typedef decltype(TimestampClock::now()) TimestampInstant;

#define __PROFILE_TIMESTAMP_ID(x,y) x##y
#define PROFILE_TIMESTAMP_ID(x,y) __PROFILE_TIMESTAMP_ID(x,y)
#define PROFILE_START(stack, target) Timestamp PROFILE_TIMESTAMP_ID(timestamp_,__LINE__) (stack, target)
#define PROFILE_END(stack) (stack).top()->dispose()

extern TimestampUnit profilingOverhead;

struct ProfilingStats
{
  TimestampUnit totalTime;
  TimestampUnit maxTime;
  uint64_t count;

  void accumulate(TimestampUnit time)
  {
    totalTime += time;
    maxTime = std::max(maxTime, time);
    count++;
  }

  double totalTimeToSeconds()
  {
    return totalTime.count() / 1000000000.0;
  }

  double maxTimeToSeconds()
  {
    return maxTime.count() / 1000000000.0;
  }

  std::string toString()
  {
    std::ostringstream oss;
    oss << "Total " << totalTimeToSeconds() << "s Max " << maxTimeToSeconds() << "s Mean " << (totalTimeToSeconds() / count) << "s";
    return oss.str();
  };
};

struct Timestamp
{
  Timestamp(TimestampStack &_stack, ProfilingStats &_target)
    : target(_target), stack(_stack), startTime(TimestampClock::now())
  {
    stack.push(this);
  }

  ~Timestamp()
  {
    dispose();
  }

  void dispose()
  {
    if (stack.empty() || stack.top() != this)
      return;
    auto diff = TimestampClock::now() - startTime - profilingOverhead;
    if (diff >= TimestampUnit::zero())
      target.accumulate(diff);
    stack.pop();
  }

  ProfilingStats &target;
  TimestampStack &stack;
  TimestampInstant startTime;
};

#else

#define PROFILE_START(stack, target)
#define PROFILE_END(stack)

#endif
