#pragma once
// vim: set ts=2 sw=2 et:

#include <cstdint>
#include <type_traits>
#include <memory.h>

template<typename T>
struct PtrHashMap
{
  struct Pair
  {
    uint64_t key;
    void *value;
  };

  PtrHashMap(size_t initialCapacity) // MUST BE POWER OF 2
  {
    buckets = new Pair[initialCapacity];
    memset(buckets, 0, sizeof(Pair) * initialCapacity);
    bucketsEnd = buckets + initialCapacity;

    count = 0;
    capacity = initialCapacity;
  }

  ~PtrHashMap()
  {
    delete[] buckets;
  }

  // 60%
  const static uint64_t LOAD_FACTOR_NUM = 3;
  const static uint64_t LOAD_FACTOR_DEN = 5;

  inline size_t getCount() { return count; }

  inline static uint32_t hash(uint64_t key)
  {
    // Copied from https://code.google.com/p/fast-hash/'s mix_fasthash macro
    key ^= key >> 23;
    key *= 0x2127599bf4325c37ULL;
    key ^= key >> 47;
    return (uint32_t) key;
  }

  inline Pair *__get(uint64_t key)
  {
    uint32_t hkey = hash(key) & (capacity - 1);
    Pair *ptr = buckets + hkey;
    for (;ptr != bucketsEnd;ptr++)
    {
      if (__builtin_expect(ptr->key == key, 1) || __builtin_expect(ptr->value == nullptr, 1))
        return ptr;
    }
    for (ptr = buckets;;ptr++)
    {
      if (__builtin_expect(ptr->key == key, 1) || __builtin_expect(ptr->value == nullptr, 1))
        return ptr;
    }
  }

  inline typename std::enable_if<std::is_pointer<T>::value, T>::type get(uint64_t key)
  {
    void *v = __get(key)->value;
    if (v == buckets)
      v = nullptr;
    return reinterpret_cast<T>(v);
  }

  inline typename std::enable_if<std::is_pointer<T>::value, T>::type getAndRemove(uint64_t key)
  {
    Pair *pair = __get(key);
    void *ret = pair->value;
    if (ret == buckets || ret == nullptr)
      return nullptr;

    pair->value = buckets; // delete marker
    count--;
    return reinterpret_cast<T>(ret);
  }

  inline void set(uint64_t key, T ptr)
  {
    if (dirtyCount * LOAD_FACTOR_DEN >= capacity * LOAD_FACTOR_NUM)
      __rehash();
    Pair *pair = __get(key);
    count++;
    dirtyCount += (pair->value == nullptr);
    pair->key = key;
    pair->value = ptr;
  }

  void __rehash()
  {
    Pair *oldBuckets = buckets;
    Pair *oldBucketsEnd = bucketsEnd;

    capacity <<= 2;
    buckets = new Pair[capacity];
    memset(buckets, 0, sizeof(Pair) * capacity);
    bucketsEnd = buckets + capacity;
    count = 0;
    dirtyCount = 0;

    for (Pair *ptr = oldBuckets;ptr != oldBucketsEnd;ptr++)
      if (__builtin_expect(ptr->value != nullptr, 1) && __builtin_expect(ptr->value != oldBuckets, 1))
        set(ptr->key, reinterpret_cast<T>(ptr->value));

    delete[] oldBuckets;
  }

  Pair *buckets;
  Pair *bucketsEnd;
  size_t capacity;
  size_t count;
  size_t dirtyCount;
};
